package be.iminds.ilabt.jfed.git;

import be.iminds.ilabt.jfed.util.IOUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.file.Files;
import java.nio.file.attribute.PosixFilePermissions;

/**
 * GitFetcher: downloads git repo's locally
 */
public class GitFetcher {
    private static final Logger LOG = LoggerFactory.getLogger(GitFetcher.class);

    @Nullable private final GitAuthPreferences gitAuthPreferences;

    /**
     * @param gitAuthPreferences GitAuthPreferences to use as fallback, when no gitAuth is specified in the command itself
     */
    public GitFetcher(@Nullable GitAuthPreferences gitAuthPreferences) {
        this.gitAuthPreferences = gitAuthPreferences;
    }

    /**
     * @param gitFetcherCommand
     * @return the dir in which the files from git are downloaded. If subDir is requested, that local dir corresponding to that subDir is returned.
     */
    public File fetch(GitFetcherCommand gitFetcherCommand) {
        String subDir = gitFetcherCommand.getRepoSubDir();
        String gitUrl = gitFetcherCommand.getGitUrl();
        String branch = gitFetcherCommand.getBranch();

        File resDir;
        File rootGitDir = gitFetcherCommand.getTargetDir();
        assert rootGitDir != null;
        if (subDir != null && !subDir.trim().equals("/")) {
            resDir = new File(rootGitDir, subDir);
        } else {
            resDir = rootGitDir;
        }

        String fetchScript = "#!/bin/bash -e\n\n";
        fetchScript += "cd '"+rootGitDir+"'\n";
        fetchScript += "git init\n";
        if (subDir != null)
            fetchScript += "git config core.sparseCheckout true\n" +
                    "echo \""+subDir+"*\" >> .git/info/sparse-checkout\n";
        if (gitFetcherCommand.getGitAuth() != null) {

            GitAuthPreferences.GitAuthSsh sshGitAuth = null;

            if (gitFetcherCommand.getGitAuth() != null &&
                    gitFetcherCommand.getGitAuth().getMethod() == GitAuthPreferences.GitAuthMethod.SSH) {
                sshGitAuth = (GitAuthPreferences.GitAuthSsh) gitFetcherCommand.getGitAuth();
            }

            if (sshGitAuth == null && gitAuthPreferences != null) {
                //try fallback to gitAuthPreferences
                sshGitAuth = gitAuthPreferences.getGitAuthSsh(gitUrl);
            }

            if (sshGitAuth != null) {
                //setup SSH agent
                fetchScript += "eval $(ssh-agent -s)\n";

                //add key to ssh agent
//            for (String pem : gitFetcherCommand.getPrivateKeyPems()) {
                String pem = sshGitAuth.getPrivateKey();
                fetchScript += "KEY_FILE=$(mktemp /tmp/gfetch.XXXXXXX)\n";
                fetchScript += "cat > \"${KEY_FILE}\" <<EOF\n";
                fetchScript += "" + pem + "\nEOF\n";
                fetchScript += "ssh-add \"${KEY_FILE}\"\n";
                fetchScript += "rm \"${KEY_FILE}\"\n";
                fetchScript += "\n";
//            }
                fetchScript += "ssh-add -L\n";
                fetchScript += "\n";
            }
        }
        fetchScript += "git remote add -f origin "+gitUrl+"\n";
        fetchScript += "git pull --depth=1 origin "+(branch == null ? "master" : branch)+"\n";
        fetchScript += "echo \n";
        if (subDir != null)
            fetchScript +=  "echo 'Files:'\n"+"ls \""+subDir+"\"\n";

        File tmpScript = null;
        try {
            tmpScript = Files.createTempFile("git-checkout", ".sh",
                    PosixFilePermissions.asFileAttribute(PosixFilePermissions.fromString("rwx------")))
                    .toFile();
            IOUtils.stringToFile(tmpScript, fetchScript);
            tmpScript.deleteOnExit();

            ProcessBuilder pb = new ProcessBuilder(tmpScript.getPath());
            //noinspection ResultOfMethodCallIgnored
            pb = pb.redirectErrorStream(true);

            StringBuilder res = new StringBuilder();

            LOG.debug("Will execute git fetch with script:\n"+fetchScript);
            Process p;
            try {
                p = pb.start();
                p.getOutputStream().close(); //we don't write to the command's stdin
                BufferedReader input = new BufferedReader(new InputStreamReader(p.getInputStream()));

                String line = null;
                while ((line = input.readLine()) != null) {
                    res.append(line+"\n");
                }
                //note: input might close before terminal stops.
                input.close();
            } catch (IOException e) {
                LOG.debug("Fetch script failure:\n"+fetchScript, true);
                LOG.debug("Fetch script output:\n"+res.toString(), true);
                LOG.error("Error executing command: "+e.getMessage(), e);
                throw new RuntimeException("Exception executing fetch git commands.\n"+fetchScript, e);
            }

            try {
                p.waitFor();
            } catch (InterruptedException e) {
                LOG.warn("fetch git command thread interrupted while waiting for command to stop", e);
            }

            LOG.debug("Fetched repo from git. Output:");
            LOG.debug(res.toString());

            int exitVal = p.exitValue();
            LOG.debug("exitVal: "+exitVal);
            if (exitVal != 0) {
                LOG.debug("Fetch script exited with value"+exitVal+": "+fetchScript, true);
                throw new RuntimeException("Exit value "+exitVal+" when fetching git commands.\n");
            }

            return resDir;
        } catch (IOException e) {
            throw new RuntimeException("Error executing git fetch", e);
        } finally {
            if (tmpScript != null)
                //noinspection ResultOfMethodCallIgnored
                tmpScript.delete();
        }
    }
}
